/* -*- c-basic-offset: 8 -*-
   rdesktop: A Remote Desktop Protocol client.
   Entrypoint and utility functions
   Copyright (C) Matthew Chapman 1999-2008

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdarg.h>		/* va_list va_start va_end */
#include <unistd.h>		/* read close getuid getgid getpid getppid gethostname */
#include <getopt.h>
#include <fcntl.h>		/* open */
#include <pwd.h>		/* getpwuid */
#include <termios.h>		/* tcgetattr tcsetattr */
#include <sys/stat.h>		/* stat */
#include <sys/time.h>		/* gettimeofday */
#include <sys/times.h>		/* times */
#include <ctype.h>		/* toupper */
#include <errno.h>
#ifdef WITH_PLUGIN
#include <signal.h>		/* kill */
#endif
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#ifdef WITH_XINERAMA
#include <X11/extensions/Xinerama.h>
#endif /* WITH_XINERAMA */
#include "rdesktop.h"

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif
#ifdef HAVE_ICONV
#ifdef HAVE_LANGINFO_H
#include <langinfo.h>
#endif
#endif

#ifdef EGD_SOCKET
#include <sys/types.h>
#include <sys/socket.h>		/* socket connect */
#include <sys/un.h>		/* sockaddr_un */
#endif

#ifdef WITH_PLUGIN
#include <dlfcn.h>
#endif

#include "ssl.h"

#ifdef WITH_FABULATECH
#include "ftusbrdp.h"
#endif /* WITH_FABULATECH */

char g_title[64] = "";
char g_wmclass[64] = "";
char g_username[64];
char g_hostname[16];
char g_keymapname[PATH_MAX] = "";
unsigned int g_keylayout = 0x409;	/* Defaults to US keyboard layout */
int g_keyboard_type = 0x4;	/* Defaults to US keyboard layout */
int g_keyboard_subtype = 0x0;	/* Defaults to US keyboard layout */
int g_keyboard_functionkeys = 0xc;	/* Defaults to US keyboard layout */

int g_width = 800;		/* width is special: If 0, the
				   geometry will be fetched from
				   _NET_WORKAREA. If negative,
				   absolute value specifies the
				   percent of the whole screen. */
int g_height = 600;
#ifdef WITH_XINERAMA
int g_x = -1;
int g_y = -1;
int g_x_wait = -1;
int g_y_wait = -1;
int g_percentage_width = -1;
int g_percentage_height = -1;
#endif /* WITH_XINERAMA */
int g_xpos = 0;
int g_ypos = 0;
int g_pos = 0;			/* 0 position unspecified,
				   1 specified,
				   2 xpos neg,
				   4 ypos neg  */
extern int g_tcp_port_rdp;
extern Display *g_display;
int g_server_depth = -1;
int g_win_button_size = 0;	/* If zero, disable single app mode */
int g_use_rdp = RDP5;
RD_BOOL g_bitmap_compression = True;
RD_BOOL g_sendmotion = True;
RD_BOOL g_bitmap_cache = True;
RD_BOOL g_bitmap_cache_persist_enable = False;
RD_BOOL g_bitmap_cache_precache = True;
RD_BOOL g_encryption = True;
RD_BOOL g_packet_encryption = True;
RD_BOOL g_desktop_save = True;	/* desktop save order */
RD_BOOL g_polygon_ellipse_orders = True;	/* polygon / ellipse orders */
RD_BOOL g_fullscreen = False;
RD_BOOL g_geometry = False;
RD_BOOL g_grab_keyboard = True;
RD_BOOL g_hide_decorations = False;
RD_BOOL g_rdpclip = True;
RD_BOOL g_console_session = False;
RD_BOOL g_numlock_sync = False;
RD_BOOL g_lspci_enabled = False;
RD_BOOL g_owncolmap = False;
RD_BOOL g_ownbackstore = True;	/* We can't rely on external BackingStore */
RD_BOOL g_seamless_rdp = False;
RD_BOOL g_allow_toggle = False;
uint32 g_embed_wnd;
uint32 g_rdp5_performanceflags =
	RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG | RDP5_NO_MENUANIMATIONS;
/* Session Directory redirection */
RD_BOOL g_redirect = False;
char g_redirect_server[64];
char g_redirect_domain[16];
ByteArray g_redirect_password;
char g_redirect_username[64];
char g_redirect_cookie[128];
uint32 g_redirect_session_id = 0;
uint32 g_redirect_flags = 0;

#ifdef WITH_RDPSND
RD_BOOL g_rdpsnd = False;
#endif

#ifdef WITH_FABULATECH
RD_BOOL g_ftusbrdp = False;
#endif /* WITH_FABULATECH */

#ifdef HAVE_ICONV
char g_codepage[16] = "";
#endif

#ifdef WITH_PLUGIN
PLUGIN *plugins = NULL;
ADDIN_DATA addin_data[MAX_ADDINS];
int addin_count = 0;
#ifdef WITH_WYSE_MMR
RD_BOOL g_notify = False;
#endif
#endif

#ifdef WITH_XINERAMA
RD_BOOL g_use_xinerama = False;
#endif /* WITH_XINERAMA */

extern RDPDR_DEVICE g_rdpdr_device[];
extern uint32 g_num_devices;
extern char *g_rdpdr_clientname;
extern Screen *g_screen;
extern int g_depth;
extern Visual *g_visual;
extern int g_shutdown;
extern int g_is_connected;
extern RD_BOOL g_licence_issued;

#ifdef WITH_XINERAMA
extern XineramaScreenInfo *g_xinerama_screens;
extern int g_xinerama_num_screens;
#endif /* WITH_XINERAMA */

#ifdef WITH_PLUGIN
extern Window g_wnd;
#endif

#ifdef RDP2VNC
extern int rfb_port;
extern int defer_time;
void
rdp2vnc_connect(char *server, uint32 flags, char *domain, ByteArray *password,
		char *shell, char *directory);
#endif

#if defined(WITH_PLUGIN) && defined(WITH_WYSE_MMR)
int
load_wyse_plugins(char *module);
Window
get_rdp_window();
#endif

#if defined(WITH_PLUGIN) && defined(WITH_WYSE_MMR)
static RDP_SESSION rdp_session =
{
	get_rdp_window,
	channel_old_register,
	channel_init,
	channel_send,
	channel_process
};
#endif

/* Display usage information */
static void
usage(char *program)
{
	fprintf(stderr, "rdesktop: A Remote Desktop Protocol client.\n");
	fprintf(stderr, "Version " VERSION ". Copyright (C) 1999-2008 Matthew Chapman.\n");
	fprintf(stderr, "See http://www.rdesktop.org/ for more information.\n\n");

	fprintf(stderr, "Usage: %s [options] server[:port]\n", program);
#ifdef RDP2VNC
	fprintf(stderr, "   -V: vnc port\n");
	fprintf(stderr, "   -Q: defer time (ms)\n");
#endif
	fprintf(stderr, "   -u: user name\n");
	fprintf(stderr, "   -d: domain\n");
	fprintf(stderr, "   -s: shell\n");
	fprintf(stderr, "   -c: working directory\n");
	fprintf(stderr, "   -p: password (- to prompt)\n");
	fprintf(stderr, "   -n: client hostname\n");
	fprintf(stderr, "   -k: keyboard layout on server (en-us, de, sv, etc.)\n");
	fprintf(stderr, "   -g: desktop geometry (WxH)\n");
#ifdef WITH_XINERAMA
	fprintf(stderr, "   -M: select view screen in window mode\n");
	fprintf(stderr, "   -F: full-screen over specified screens (inherits -D and -K)\n");
#endif /* WITH_XINERAMA */
	fprintf(stderr, "   -f: full-screen mode\n");
	fprintf(stderr, "   -b: force bitmap updates\n");
#ifdef HAVE_ICONV
	fprintf(stderr, "   -L: local codepage\n");
#endif
	fprintf(stderr, "   -A: enable SeamlessRDP mode\n");
	fprintf(stderr, "   -B: use BackingStore of X-server (if available)\n");
	fprintf(stderr, "   -e: disable encryption (French TS)\n");
	fprintf(stderr, "   -E: disable encryption from client to server\n");
	fprintf(stderr, "   -m: do not send motion events\n");
	fprintf(stderr, "   -C: use private colour map\n");
	fprintf(stderr, "   -D: hide window manager decorations\n");
	fprintf(stderr, "   -K: keep window manager key bindings\n");
	fprintf(stderr, "   -S: caption button size (single application mode)\n");
	fprintf(stderr, "   -T: window title\n");
	fprintf(stderr, "   -W: window class\n");
	fprintf(stderr, "   -N: enable numlock syncronization\n");
	fprintf(stderr, "   -X: embed into another window with a given id.\n");
	fprintf(stderr, "   -a: connection colour depth\n");
	fprintf(stderr, "   -z: enable rdp compression\n");
	fprintf(stderr, "   -x: RDP5 experience (m[odem 28.8], b[roadband], l[an] or hex nr.)\n");
	fprintf(stderr, "   -P: use persistent bitmap caching\n");
	fprintf(stderr, "   -r: enable specified device redirection (this flag can be repeated)\n");
	fprintf(stderr,
		"         '-r comport:COM1=/dev/ttyS0': enable serial redirection of /dev/ttyS0 to COM1\n");
	fprintf(stderr, "             or      COM1=/dev/ttyS0,COM2=/dev/ttyS1\n");
	fprintf(stderr,
		"         '-r disk:floppy=/mnt/floppy': enable redirection of /mnt/floppy to 'floppy' share\n");
#ifdef WITH_PLUGIN
	fprintf(stderr,
		"         '-r plugin:type:/path/to/module.so[:arg0[,arg1...]]'\n");
	fprintf(stderr,
		"             'type' is one of the following strings:\n");
	fprintf(stderr,
		"             elusiva:      Support for third party virtual channel add-ins which uses the elusiva interface\n");
	fprintf(stderr,
		"                               arg0 must be the channel name\n");
#ifdef WITH_WYSE_MMR
	fprintf(stderr,
		"             wyse_mmr:     Support for Wyse Multimedia Redirection\n");
#endif
#endif
	fprintf(stderr, "             or   'floppy=/mnt/floppy,cdrom=/mnt/cdrom'\n");
	fprintf(stderr, "         '-r clientname=<client name>': Set the client name displayed\n");
	fprintf(stderr, "             for redirected disks\n");
	fprintf(stderr,
		"         '-r lptport:LPT1=/dev/lp0': enable parallel redirection of /dev/lp0 to LPT1\n");
	fprintf(stderr, "             or      LPT1=/dev/lp0,LPT2=/dev/lp1\n");
	fprintf(stderr, "         '-r printer:mydeskjet': enable printer redirection\n");
	fprintf(stderr,
		"             or      mydeskjet=\"HP LaserJet IIIP\" to enter server driver as well\n");
#ifdef WITH_RDPSND
	fprintf(stderr,
		"         '-r sound:[local[:driver[:device]]|off|remote]': enable sound redirection\n");
	fprintf(stderr, "                     remote would leave sound on server\n");
	fprintf(stderr, "                     available drivers for 'local':\n");
	rdpsnd_show_help();
#endif
	fprintf(stderr,
		"         '-r clipboard:[off|PRIMARYCLIPBOARD|CLIPBOARD]': enable clipboard\n");
	fprintf(stderr, "                      redirection.\n");
	fprintf(stderr,
		"                      'PRIMARYCLIPBOARD' looks at both PRIMARY and CLIPBOARD\n");
	fprintf(stderr, "                      when sending data to server.\n");
	fprintf(stderr, "                      'CLIPBOARD' looks at only CLIPBOARD.\n");
#ifdef WITH_SCARD
	fprintf(stderr, "         '-r scard[:\"Scard Name\"=\"Alias Name[;Vendor Name]\"[,...]]\n");
	fprintf(stderr, "          example: -r scard:\"eToken PRO 00 00\"=\"AKS ifdh 0\"\n");
	fprintf(stderr,
		"                   \"eToken PRO 00 00\" -> Device in Linux/Unix enviroment\n");
	fprintf(stderr,
		"                   \"AKS ifdh 0\"       -> Device shown in Windows enviroment \n");
	fprintf(stderr, "          example: -r scard:\"eToken PRO 00 00\"=\"AKS ifdh 0;AKS\"\n");
	fprintf(stderr,
		"                   \"eToken PRO 00 00\" -> Device in Linux/Unix enviroment\n");
	fprintf(stderr,
		"                   \"AKS ifdh 0\"       -> Device shown in Windows enviroment \n");
	fprintf(stderr,
		"                   \"AKS\"              -> Device vendor name                 \n");
#endif
	fprintf(stderr, "   -0: attach to console\n");
	fprintf(stderr, "   -4: use RDP version 4\n");
	fprintf(stderr, "   -5: use RDP version 5 (default)\n");
#ifdef USE_RDP6
	fprintf(stderr, "   -6: use RDP version 6\n");
#endif /* USE_RDP6 */
	fprintf(stderr, "   --allow-toggle:     allow to toggle between fullscreen "
					"and window mode\n");
}

static void
print_disconnect_reason(uint16 reason)
{
	char *text;

	switch (reason)
	{
		case exDiscReasonInitiatedDisconnectByUser:
			text = "Disconnection was initiated by an administrative tool";
			break;

		case exDiscReasonNoInfo:
			text = "No information available";
			break;

		case exDiscReasonAPIInitiatedDisconnect:
			text = "Server initiated disconnect";
			break;

		case exDiscReasonAPIInitiatedLogoff:
			text = "Server initiated logoff";
			break;

		case exDiscReasonServerIdleTimeout:
			text = "Server idle timeout reached";
			break;

		case exDiscReasonServerLogonTimeout:
			text = "Server logon timeout reached";
			break;

		case exDiscReasonReplacedByOtherConnection:
			text = "The session was replaced";
			break;

		case exDiscReasonOutOfMemory:
			text = "The server is out of memory";
			break;

		case exDiscReasonServerDeniedConnection:
			text = "The server denied the connection";
			break;

		case exDiscReasonServerDeniedConnectionFips:
			text = "The server denied the connection for security reason";
			break;

		case exDiscReasonLicenseInternal:
			text = "Internal licensing error";
			break;

		case exDiscReasonLicenseNoLicenseServer:
			text = "No license server available";
			break;

		case exDiscReasonLicenseNoLicense:
			text = "No valid license available";
			break;

		case exDiscReasonLicenseErrClientMsg:
			text = "Invalid licensing message";
			break;

		case exDiscReasonLicenseHwidDoesntMatchLicense:
			text = "Hardware id doesn't match software license";
			break;

		case exDiscReasonLicenseErrClientLicense:
			text = "Client license error";
			break;

		case exDiscReasonLicenseCantFinishProtocol:
			text = "Network error during licensing protocol";
			break;

		case exDiscReasonLicenseClientEndedProtocol:
			text = "Licensing protocol was not completed";
			break;

		case exDiscReasonLicenseErrClientEncryption:
			text = "Incorrect client license enryption";
			break;

		case exDiscReasonLicenseCantUpgradeLicense:
			text = "Can't upgrade license";
			break;

		case exDiscReasonLicenseNoRemoteConnections:
			text = "The server is not licensed to accept remote connections";
			break;

		default:
			if (reason > 0x1000 && reason < 0x7fff)
			{
				text = "Internal protocol error";
			}
			else
			{
				text = "Unknown reason";
			}
			fprintf(stderr, "disconnect: %s: 0x%X.\n", text, reason);
			return;
	}
	fprintf(stderr, "disconnect: %s.\n", text);
}

static void
rdesktop_reset_state(void)
{
	rdp_reset_state();
}

static RD_BOOL
read_password(ByteArray *password)
{
	struct termios tios;
	RD_BOOL ret = False;
	int istty = 0;
	char *p;

	if (tcgetattr(STDIN_FILENO, &tios) == 0)
	{
		fprintf(stdout, "Password: ");
		tios.c_lflag &= ~ECHO;
		tcsetattr(STDIN_FILENO, TCSANOW, &tios);
		istty = 1;
	}

	password->bytes = malloc(sizeof(char) * 64);
	if (fgets(password->bytes, 64, stdin) != NULL)
	{
		ret = True;

		/* strip final newline */
		p = strchr(password->bytes, '\n');
		if (p != NULL)
			*p = 0;
		password->len = strlen(password->bytes);
		password->toUTF16 = True;
	}

	if (istty)
	{
		tios.c_lflag |= ECHO;
		tcsetattr(STDIN_FILENO, TCSANOW, &tios);
		fprintf(stdout, "\n");
	}

	return ret;
}

static void
parse_server_and_port(char *server)
{
	char *p;
#ifdef IPv6
	int addr_colons;
#endif

#ifdef IPv6
	p = server;
	addr_colons = 0;
	while (*p)
		if (*p++ == ':')
			addr_colons++;
	if (addr_colons >= 2)
	{
		/* numeric IPv6 style address format - [1:2:3::4]:port */
		p = strchr(server, ']');
		if (*server == '[' && p != NULL)
		{
			if (*(p + 1) == ':' && *(p + 2) != '\0')
				g_tcp_port_rdp = strtol(p + 2, NULL, 10);
			/* remove the port number and brackets from the address */
			*p = '\0';
			strncpy(server, server + 1, strlen(server));
		}
	}
	else
	{
		/* dns name or IPv4 style address format - server.example.com:port or 1.2.3.4:port */
		p = strchr(server, ':');
		if (p != NULL)
		{
			g_tcp_port_rdp = strtol(p + 1, NULL, 10);
			*p = 0;
		}
	}
#else /* no IPv6 support */
	p = strchr(server, ':');
	if (p != NULL)
	{
		g_tcp_port_rdp = strtol(p + 1, NULL, 10);
		*p = 0;
	}
#endif /* IPv6 */

}

static void
print_wait_message(char* server, Window window, unsigned long white, unsigned long black)
{
	GC gc;
	char* message;
	char mask[]= "7x14";
	Font font;
	XFontStruct* font_info;
	int width= 0;
	int height;

	message= xmalloc(strlen(server)+ strlen(WAIT_WND_MSG)+ 4);
	sprintf(message, "%s%s", WAIT_WND_MSG, server);

	font= XLoadFont(g_display, mask);
	font_info= XQueryFont(g_display, font);
	if ((width= XTextWidth(font_info, message, strlen(message))) >
		WAIT_WND_WIDTH- 10)
	{
		sprintf(message, "%s...", message);
		while (((width= XTextWidth(font_info, message, strlen(message))) >
			WAIT_WND_WIDTH- 10) && (strlen(message) > 5))
		{
			message[strlen(message)- 4]= '.';
			message[strlen(message)- 1]= '\0';
		}
	}
	height= font_info->ascent;

	gc= XCreateGC(g_display, window, 0, 0);
	XSetFont(g_display, gc, font);

	XSetForeground(g_display, gc, black);
	XSetBackground(g_display, gc, white);

	XDrawString(g_display, window, gc, (WAIT_WND_WIDTH/ 2)- (width/ 2),
		(WAIT_WND_HEIGHT- height)/ 2+ height, message, strlen(message));

	XUnloadFont(g_display, font);
	XFreeFontInfo(NULL, font_info, 0);
	XFreeGC(g_display, gc);
	xfree(message);
}

static Window
show_wait_window(char *server)
{
	int white;
	int black;
	Window window;
	XClassHint class_hints;
	XSetWindowAttributes attr;
	unsigned long x;
	unsigned long y;


	white= WhitePixel(g_display, DefaultScreen(g_display));
	black= BlackPixel(g_display, DefaultScreen(g_display));
	x= (WidthOfScreen(g_screen) - WAIT_WND_WIDTH) / 2;
	y= (HeightOfScreen(g_screen) - WAIT_WND_HEIGHT) / 2;
#ifdef WITH_XINERAMA
	if (g_x_wait >= 0)
	{
		x = g_x_wait;
		y = g_y_wait;
	}
#endif /* WITH_XINERAMA */

	attr.background_pixel= white;
	attr.border_pixel= black;
	attr.override_redirect= True;

	window= XCreateWindow( g_display, RootWindowOfScreen(g_screen), x, y,
		WAIT_WND_WIDTH, WAIT_WND_HEIGHT, 2, g_depth, InputOutput,
		CopyFromParent, CWBackPixel | CWOverrideRedirect | CWBorderPixel,
		&attr);

	if (strlen(g_wmclass) > 0)
		class_hints.res_name= g_wmclass;
	else
		class_hints.res_name= "rdesktop";
	class_hints.res_class= "rdesktop-connect";
	XSetClassHint(g_display, window, &class_hints);

	XMapWindow(g_display, window);
	print_wait_message(server, window, white, black);
	XFlush(g_display);

	return window;
}

static void
hide_wait_window(Window window)
{
	XUnmapWindow(g_display, window);
	XDestroyWindow(g_display, window);
}

#ifdef WITH_PLUGIN
static int
parse_plugin(char *str)
{
	int ret = False;
	int type = NO_PLUGIN;
	int i;
	char *p;
	char *module = NULL;
	char **args = NULL;

	if (*str == ':')
	{
		/* Check if plugin is supported */
		str++;
		p = next_arg(str, ':');
		if ((str == NULL) || (strlen(str) == 0))
			error("Missing type of plugin\n");
		else if (strcmp( str, "elusiva") == 0)
		{
			type = ELUSIVA;
			addin_count++;
		}
#ifdef WITH_WYSE_MMR
		else if (strcmp(str, "wyse_mmr") == 0)
			type = WYSE_MMR;
#endif
		else
			error("Type of plugin unknown\n");

		if (type == NO_PLUGIN)
			return False;

		if ((type == ELUSIVA) && (addin_count >= MAX_ADDINS))
		{
			error("Too much elusiva plugins\n");
			return False;
		}

		/* Parse module name */
		str = p;
		p = next_arg(str, ':');
		if ((str == NULL) || (strlen(str) == 0))
			error("Module name is missing\n");
		else
			module = xstrdup(str);

		if (module == NULL)
			return False;

		/* Save arguments */
		str = p;
		while ((p = next_arg(str, ',')))
		{
			if (args == NULL) {
				args = xmalloc(sizeof(char *) * 2);
				args[0] = xstrdup(str);
				args[1] = NULL;
			}
			else
			{
				i = 0;
				while (args[i] != NULL)
					i++;
				args = xrealloc(args, sizeof(char *) * (i + 2));
				args[i + 1] = xstrdup(str);
				args[i + 2] = NULL;

			}
			str = p;
		}

		if ((type == ELUSIVA) && (args == NULL))
		{
			error("Channel name of the elusiva add-in is missing\n");
			return False;
		}

		/* Save plugin data */
		if (plugins == NULL)
		{
			plugins = xmalloc(sizeof(PLUGIN) * 2);
			plugins[0].type = type;
			plugins[0].module = module;
			plugins[0].args = args;
			plugins[1].type = NO_PLUGIN;
			plugins[1].module = NULL;
			plugins[1].args = NULL;
		}
		else
		{
			i = 0;
			while (plugins[i].type != NO_PLUGIN)
				i++;
			plugins = xrealloc(plugins, sizeof(PLUGIN) * (i + 2));
			plugins[i].type = type;
			plugins[i].module = module;
			plugins[i].args = args;
			plugins[i + 1].type = NO_PLUGIN;
			plugins[i + 1].module = NULL;
			plugins[i + 1].args = NULL;
		}

		ret = True;
	}
	else
		error("Missing type of plugin\n");

	return ret;
}
#endif

#ifdef WITH_XINERAMA
static int
parse_xinerama_fullscreens(char* optarg) {
	int screen;
	int monitor = 0;
	int num_monitors = 0;
	unsigned char screen_list[16];
	char *p = NULL;
	int i, j, k;
	int right;
	int bottom;

	memset(screen_list, 0, sizeof(screen_list));

	g_display = XOpenDisplay(NULL);
	if (g_display == NULL)
	{
		error("Failed to open display: %s\n", XDisplayName(NULL));
		return 1;
	}

	/* Set defaults */
	screen = XDefaultScreen(g_display);
	g_x = 0;
	g_y = 0;
	g_width = XDisplayWidth(g_display, screen);
	g_height = XDisplayHeight(g_display, screen);

	/* Get Xinerama informations */
	if (!XineramaIsActive(g_display) ||
		((g_xinerama_screens = XineramaQueryScreens(g_display,
		&g_xinerama_num_screens)) == NULL))
	{
		warning("Xinerama is not available, using all monitors\n");
		g_use_xinerama = False;
		return 0;
	}

	/* Expand over all screens */
	if ((strtol(optarg, &p, 10) == -1) && (p != NULL) && (*p == '\0')) {
		g_x_wait = g_xinerama_screens[0].x_org +
			((g_xinerama_screens[0].width - WAIT_WND_WIDTH) / 2);
		g_y_wait = g_xinerama_screens[0].y_org +
			((g_xinerama_screens[0].height - WAIT_WND_HEIGHT) / 2);
		if (g_xinerama_num_screens == 1)
			g_use_xinerama = False;	/* don't send sizes if only on is available */
		return 0;
	}

	/* Parse argument string */
	p = optarg;
	while (*p)
	{
		monitor = strtol(p, &p, 10);
		if((monitor < 0) || (monitor >= RDP_MAX_SCREENS))
		{
			error("Invalid screen %d\n", monitor);
			return 1;
		}
		if (monitor >= g_xinerama_num_screens)
		{
			warning("Screen %d not available, using screen 0 instead\n",
				monitor);
			monitor = 0;
		}
		num_monitors++;
		screen_list[monitor] = 1;

		if((*p != '\0') && (*p != ','))
		{
			error("Invalid screen list: %s\n", optarg);
			return 1;
		}

		if(*p != '\0')
			p++;
	}

	if (num_monitors == 0)
	{
		/* Wrong configuration */
		error("No screens have been defined: %s\n", optarg);
		return 1;
	}
	else if (num_monitors == 1)
	{
		/* Only one screen, xinerama not necessary */
		g_x = g_xinerama_screens[monitor].x_org;
		g_y = g_xinerama_screens[monitor].y_org;
		g_width = g_xinerama_screens[monitor].width;
		g_height = g_xinerama_screens[monitor].height;
		g_use_xinerama = False;
		g_x_wait = g_xinerama_screens[monitor].x_org +
			((g_xinerama_screens[monitor].width - WAIT_WND_WIDTH) / 2);
		g_y_wait = g_xinerama_screens[monitor].y_org +
			((g_xinerama_screens[monitor].height - WAIT_WND_HEIGHT) / 2);
	}
	else
	{
		/* Set start values for x and y */
		g_x = g_xinerama_screens[monitor].x_org;
		g_y = g_xinerama_screens[monitor].y_org;
		right = g_xinerama_screens[monitor].x_org +
			g_xinerama_screens[monitor].width - 1;
		bottom = g_xinerama_screens[monitor].y_org +
			g_xinerama_screens[monitor].height - 1;


		/* Remove not used definitions */
		i = 0;
		j = 0;
		while (i < g_xinerama_num_screens)
		{
			if (screen_list[j]) {
				if (g_xinerama_screens[i].x_org < g_x)
					g_x = g_xinerama_screens[i].x_org;
				if (g_xinerama_screens[i].y_org < g_y)
					g_y = g_xinerama_screens[i].y_org;
				if (g_xinerama_screens[i].x_org +
					g_xinerama_screens[i].width - 1 > right)
					right = g_xinerama_screens[i].x_org +
						g_xinerama_screens[i].width - 1;
				if (g_xinerama_screens[i].y_org +
					g_xinerama_screens[i].height - 1 > bottom)
					bottom = g_xinerama_screens[i].y_org +
						g_xinerama_screens[i].height - 1;
				i++;
			}
			else
			{
				if (i < g_xinerama_num_screens - 1)
					for (k = i; k < g_xinerama_num_screens - 1; k++)
						g_xinerama_screens[k] = g_xinerama_screens[k + 1];
				g_xinerama_num_screens--;
			}
			j++;
		}

		for (i = 0; i < g_xinerama_num_screens; i++)
		{
			g_xinerama_screens[i].x_org -= g_x;
			g_xinerama_screens[i].y_org -= g_y;
		}

		g_width = (right - g_x) + 1;
		g_height = (bottom - g_y) + 1;
		g_x_wait = g_xinerama_screens[0].x_org +
			((g_xinerama_screens[0].width - WAIT_WND_WIDTH) / 2);
		g_y_wait = g_xinerama_screens[0].y_org +
			((g_xinerama_screens[0].height - WAIT_WND_HEIGHT) / 2);
	}

	return 0;
}

static int
parse_xinerama_screens(char* optarg) {
	int screen;
	int monitor = 0;
	char *p = NULL;

	g_display = XOpenDisplay(NULL);
	if (g_display == NULL)
	{
		error("Failed to open display: %s\n", XDisplayName(NULL));
		return 1;
	}

	/* Set defaults */
	screen = XDefaultScreen(g_display);
	g_x = 0;
	g_y = 0;
	g_percentage_width = XDisplayWidth(g_display, screen);
	g_percentage_height = XDisplayHeight(g_display, screen);
	g_x_wait = (g_percentage_width - WAIT_WND_WIDTH) / 2;
	g_y_wait = (g_percentage_height - WAIT_WND_HEIGHT) / 2;

	/* Get Xinerama informations */
	if (!XineramaIsActive(g_display) ||
		((g_xinerama_screens = XineramaQueryScreens(g_display,
		&g_xinerama_num_screens)) == NULL))
	{
		warning("Xinerama is not available, using all monitors\n");
		return 0;
	}

	/* Expand over all screens */
	if ((strtol(optarg, &p, 10) == -1) && (p != NULL) && (*p == '\0')) {
		if (g_xinerama_num_screens != 1)
		{
			g_x_wait = g_xinerama_screens[0].x_org + 
				((g_xinerama_screens[0].width - WAIT_WND_WIDTH) / 2);
			g_y_wait = g_xinerama_screens[0].y_org +
				((g_xinerama_screens[0].height - WAIT_WND_HEIGHT) / 2);
		}
		return 0;
	}

	/* Parse argument string */
	monitor = strtol(optarg, &p, 10);
	if ((monitor < 0) || (monitor >= RDP_MAX_SCREENS) || (p == NULL) ||
		(*p != '\0'))
	{
		error("Invalid screen definition: %s\n", optarg);
		return 1;
	}
	else
	{
		if (monitor > g_xinerama_num_screens)
		{
			warning("Screen %d not available, using screen 0 instead\n",
				monitor);
			monitor = 0;
		}
		g_x = g_xinerama_screens[monitor].x_org;
		g_y = g_xinerama_screens[monitor].y_org;
		g_x_wait = g_xinerama_screens[monitor].x_org + 
			((g_xinerama_screens[monitor].width - WAIT_WND_WIDTH) / 2);
		g_y_wait = g_xinerama_screens[monitor].y_org +
			((g_xinerama_screens[monitor].height - WAIT_WND_HEIGHT) / 2);
		g_percentage_width = g_xinerama_screens[monitor].width;
		g_percentage_height = g_xinerama_screens[monitor].height;
	}

	return 0;
}
#endif /* WITH_XINERAMA */

/* Client program */
int
main(int argc, char *argv[])
{
	char server[64];
	char fullhostname[64];
	char domain[16];
	ByteArray password;
	char shell[256];
	char directory[256];
	RD_BOOL prompt_password, deactivated;
	/* struct passwd *pw; */
	uint32 flags, ext_disc_reason = 0;
	char *p;
	char *args;
	int i;
	int c;
	int eluc;
	int ret = 0;
	char *locale = NULL;
	int username_option = 0;
	RD_BOOL geometry_option = False;
	int run_count = 0;	/* Session Directory support */
	RD_BOOL continue_connect = True;	/* Session Directory support */
	Window window;
#ifdef WITH_RDPSND
	char *rdpsnd_optarg = NULL;
#endif
#ifdef WITH_XINERAMA
	RD_BOOL param_M_used = False;
	RD_BOOL param_F_used = False;
#endif /* WITH_XINERAMA */
	/* List of long options */
	static struct option longopts[]=
	{
		{"allow-toggle", no_argument, &g_allow_toggle, True},
		{NULL, 0, NULL, 0}	/* End of list */
	};

#ifdef HAVE_LOCALE_H
	/* Set locale according to environment */
	locale = setlocale(LC_ALL, "");
	if (locale)
	{
		locale = xstrdup(locale);
	}

#endif
	flags = RDP_LOGON_NORMAL;
	prompt_password = False;
	domain[0] = shell[0] = directory[0] = 0;
	password.len = 0;
	password.bytes = NULL;
	password.toUTF16 = False;
	g_embed_wnd = 0;

	g_num_devices = 0;

#ifdef RDP2VNC
#define VNCOPT "V:Q:"
#else
#define VNCOPT
#endif

	while ((c = getopt_long(argc, argv,
			   VNCOPT "Au:L:d:s:c:p:n:k:F:M:g:fbBeEmzCDKS:T:W:NX:a:x:Pr:0456h?",
			   longopts, NULL)) != -1)
	{
		switch (c)
		{
			case 0:	/* Value has been set directly */
				break;
#ifdef RDP2VNC
			case 'V':
				rfb_port = strtol(optarg, NULL, 10);
				if (rfb_port < 100)
					rfb_port += 5900;
				break;

			case 'Q':
				defer_time = strtol(optarg, NULL, 10);
				if (defer_time < 0)
					defer_time = 0;
				break;
#endif

			case 'A':
				g_seamless_rdp = True;
				break;

			case 'u':
				STRNCPY(g_username, optarg, sizeof(g_username));
				username_option = 1;
				break;

			case 'L':
#ifdef HAVE_ICONV
				STRNCPY(g_codepage, optarg, sizeof(g_codepage));
#else
				error("iconv support not available\n");
#endif
				break;

			case 'd':
				STRNCPY(domain, optarg, sizeof(domain));
				break;

			case 's':
				STRNCPY(shell, optarg, sizeof(shell));
				break;

			case 'c':
				STRNCPY(directory, optarg, sizeof(directory));
				break;

			case 'p':
				if ((optarg[0] == '-') && (optarg[1] == 0))
				{
					prompt_password = True;
					break;
				}
				password.len = strlen(optarg);
				password.toUTF16 = True;
				password.bytes = malloc(sizeof(char) * (strlen(optarg) + 1));
				sprintf(password.bytes, optarg);
				flags |= RDP_LOGON_AUTO;

				/* try to overwrite argument so it won't appear in ps */
				p = optarg;
				while (*p)
					*(p++) = 'X';
				break;

			case 'n':
				STRNCPY(g_hostname, optarg, sizeof(g_hostname));
				break;

			case 'k':
				STRNCPY(g_keymapname, optarg, sizeof(g_keymapname));
				break;

			case 'g':
				geometry_option = True;
				g_fullscreen = False;
				g_geometry = True;
				if (!strcmp(optarg, "workarea"))
				{
					g_width = g_height = 0;
					break;
				}

				g_width = strtol(optarg, &p, 10);
				if (g_width <= 0)
				{
					error("invalid geometry\n");
					return 1;
				}

				if (*p == 'x')
					g_height = strtol(p + 1, &p, 10);

				if (g_height <= 0)
				{
					error("invalid geometry\n");
					return 1;
				}

				if (*p == '%')
				{
					g_width = -g_width;
					p++;
				}

				if (*p == '+' || *p == '-')
				{
					g_pos |= (*p == '-') ? 2 : 1;
					g_xpos = strtol(p, &p, 10);

				}
				if (*p == '+' || *p == '-')
				{
					g_pos |= (*p == '-') ? 4 : 1;
					g_ypos = strtol(p, NULL, 10);
				}

				break;

			case 'F':
#ifdef WITH_XINERAMA
				param_F_used = True;
				g_use_xinerama = True;
				g_fullscreen = False;
				g_geometry = True;
				g_hide_decorations = True;	/* -D */
				g_grab_keyboard = False;	/* -K */

				if (parse_xinerama_fullscreens(optarg) != 0)
					return 1;
#else
				error("Xinerama support not available\n");
#endif /* WITH_XINERAMA */
				break;

			case 'M':
#ifdef WITH_XINERAMA
				param_M_used = True;
				if (parse_xinerama_screens(optarg) != 0)
					return 1;
#else
				error("Xinerama support not available\n");
#endif /* WITH_XINERAMA */
				break;

			case 'f':
				g_fullscreen = True;
				g_geometry = False;
				break;

			case 'b':
				g_bitmap_cache = False;
				break;

			case 'B':
				g_ownbackstore = False;
				break;

			case 'e':
				g_encryption = False;
				break;
			case 'E':
				g_packet_encryption = False;
				break;
			case 'm':
				g_sendmotion = False;
				break;

			case 'C':
				g_owncolmap = True;
				break;

			case 'D':
				g_hide_decorations = True;
				break;

			case 'K':
				g_grab_keyboard = False;
				break;

			case 'S':
				if (!strcmp(optarg, "standard"))
				{
					g_win_button_size = 18;
					break;
				}

				g_win_button_size = strtol(optarg, &p, 10);

				if (*p)
				{
					error("invalid button size\n");
					return 1;
				}

				break;

			case 'T':
				STRNCPY(g_title, optarg, sizeof(g_title));
				break;

			case 'W':
				STRNCPY(g_wmclass, optarg, sizeof(g_wmclass));
				break;
			
			case 'N':
				g_numlock_sync = True;
				break;

			case 'X':
				g_embed_wnd = strtol(optarg, NULL, 0);
				break;

			case 'a':
				g_server_depth = strtol(optarg, NULL, 10);
				if (g_server_depth != 8 &&
				    g_server_depth != 16 &&
				    g_server_depth != 15 && g_server_depth != 24
				    && g_server_depth != 32)
				{
					error("Invalid server colour depth.\n");
					return 1;
				}
				break;

			case 'z':
				DEBUG(("rdp compression enabled\n"));
				flags |= (RDP_LOGON_COMPRESSION | RDP_LOGON_COMPRESSION2);
				break;

			case 'x':
				if (str_startswith(optarg, "m"))	/* modem */
				{
					g_rdp5_performanceflags =
						RDP5_NO_WALLPAPER | RDP5_NO_FULLWINDOWDRAG |
						RDP5_NO_MENUANIMATIONS | RDP5_NO_THEMING;
				}
				else if (str_startswith(optarg, "b"))	/* broadband */
				{
					g_rdp5_performanceflags = RDP5_NO_WALLPAPER;
				}
				else if (str_startswith(optarg, "l"))	/* lan */
				{
					g_rdp5_performanceflags = RDP5_DISABLE_NOTHING;
				}
				else
				{
					g_rdp5_performanceflags = strtol(optarg, NULL, 16);
				}
				break;

			case 'P':
				g_bitmap_cache_persist_enable = True;
				break;

			case 'r':

				if (str_startswith(optarg, "sound"))
				{
					optarg += 5;

					if (*optarg == ':')
					{
						optarg++;
						while ((p = next_arg(optarg, ',')))
						{
							if (str_startswith(optarg, "remote"))
								flags |= RDP_LOGON_LEAVE_AUDIO;

							if (str_startswith(optarg, "local"))
#ifdef WITH_RDPSND
							{
								rdpsnd_optarg =
									next_arg(optarg, ':');
								g_rdpsnd = True;
							}

#else
								warning("Not compiled with sound support\n");
#endif

							if (str_startswith(optarg, "off"))
#ifdef WITH_RDPSND
								g_rdpsnd = False;
#else
								warning("Not compiled with sound support\n");
#endif

							optarg = p;
						}
					}
					else
					{
#ifdef WITH_RDPSND
						g_rdpsnd = True;
#else
						warning("Not compiled with sound support\n");
#endif
					}
				}
				else if (str_startswith(optarg, "plugin"))
				{
#ifdef WITH_PLUGIN
					optarg += 6;
					if (parse_plugin(optarg) == False)
						return 1;
#else
					warning("Not compiled with plugin support\n");
#endif
				}
				else if (str_startswith(optarg, "disk"))
				{
					/* -r disk:h:=/mnt/floppy */
					disk_enum_devices(&g_num_devices, optarg + 4);
				}
				else if (str_startswith(optarg, "comport"))
				{
					serial_enum_devices(&g_num_devices, optarg + 7);
				}
				else if (str_startswith(optarg, "lspci"))
				{
					g_lspci_enabled = True;
				}
				else if (str_startswith(optarg, "lptport"))
				{
					parallel_enum_devices(&g_num_devices, optarg + 7);
				}
				else if (str_startswith(optarg, "printer"))
				{
					printer_enum_devices(&g_num_devices, optarg + 7);
				}
				else if (str_startswith(optarg, "clientname"))
				{
					g_rdpdr_clientname = xmalloc(strlen(optarg + 11) + 1);
					strcpy(g_rdpdr_clientname, optarg + 11);
				}
				else if (str_startswith(optarg, "clipboard"))
				{
					optarg += 9;

					if (*optarg == ':')
					{
						optarg++;

						if (str_startswith(optarg, "off"))
							g_rdpclip = False;
						else
							cliprdr_set_mode(optarg);
					}
					else
						g_rdpclip = True;
				}
				else if (strncmp("scard", optarg, 5) == 0)
				{
#ifdef WITH_SCARD
					scard_enum_devices(&g_num_devices, optarg + 5);
#else
					warning("Not compiled with smartcard support\n");
#endif
				}
				else if (strncmp("usb", optarg, 3) == 0)
				{
#ifdef WITH_FABULATECH
					g_ftusbrdp = True;
#else
					warning("Not compiled with FabulaTech USB support\n");
#endif /* WITH_FABULATECH */
				}
				else
				{
					warning("Unknown -r argument\n\n\tPossible arguments are: comport, disk, lptport, printer, sound, clipboard, scard\n");
				}
				break;

			case '0':
				g_console_session = True;
				break;

			case '4':
				g_use_rdp = RDP4;
				break;

			case '5':
				g_use_rdp = RDP5;
				break;

			case '6':
				g_use_rdp = RDP6;
				break;

			case 'h':
			case '?':
			default:
				usage(argv[0]);
				return 1;
		}
	}

	if (argc - optind != 1)
	{
		usage(argv[0]);
		return 1;
	}

	STRNCPY(server, argv[optind], sizeof(server));
	parse_server_and_port(server);

#ifdef WITH_XINERAMA
	if (g_use_xinerama && (g_use_rdp == RDP4))
	{
		warning("RDP 4 does not support multi monitor configurations, disabling multi monitor support\n");
		g_use_xinerama = False;
	}

	if (param_F_used)
	{
		if (geometry_option)
		{
			error("You cannot use -F and -g at the same time\n");
			return 1;
		}
		if (g_fullscreen)
		{
			error("You cannot use -F and -f at the same time\n");
			return 1;
		}
		if (param_M_used)
		{
			error("You cannot use -F and -M at the same time\n");
			return 1;
		}
		if (g_embed_wnd)
		{
			error("You cannot use -F and -X at the same time\n");
			return 1;
		}
		if (strlen(g_wmclass) > 0)
		{
			error("You cannot use -F and -W at the same time\n");
			return 1;
		}
	}

	if (param_M_used)
	{
		if (g_fullscreen)
		{
			error("You cannot use -M and -f at the same time\n");
			return 1;
		}
		if (g_embed_wnd)
		{
			error("You cannot use -M and -X at the same time\n");
			return 1;
		}
		if (strlen(g_wmclass) > 0)
		{
			error("You cannot use -M and -W at the same time\n");
			return 1;
		}
	}
#endif

	if (g_seamless_rdp)
	{
#ifdef WITH_XINERAMA
		if (param_M_used)
		{
			error("You cannot use -A and -M at the same time\n");
			return 1;
		}
		if (param_F_used)
		{
			error("You cannot use -A and -F at the same time\n");
			return 1;
		}
#endif /* WITH_XINERAMA */
		if (g_win_button_size)
		{
			error("You cannot use -S and -A at the same time\n");
			return 1;
		}
		g_rdp5_performanceflags &= ~RDP5_NO_FULLWINDOWDRAG;
		if (geometry_option)
		{
			error("You cannot use -g and -A at the same time\n");
			return 1;
		}
		if (g_fullscreen)
		{
			error("You cannot use -f and -A at the same time\n");
			return 1;
		}
		if (g_hide_decorations)
		{
			error("You cannot use -D and -A at the same time\n");
			return 1;
		}
		if (g_embed_wnd)
		{
			error("You cannot use -X and -A at the same time\n");
			return 1;
		}
		if (g_use_rdp == RDP4)
		{
			error("You cannot use -4 and -A at the same time\n");
			return 1;
		}
		g_width = -100;
		g_grab_keyboard = False;
	}

	if (!username_option)
	{
		STRNCPY(g_username, "", sizeof(g_username));
        /*
		pw = getpwuid(getuid());
		if ((pw == NULL) || (pw->pw_name == NULL))
		{
			error("could not determine username, use -u\n");
			return 1;
		}

		STRNCPY(g_username, pw->pw_name, sizeof(g_username));
        */
	}

#ifdef HAVE_ICONV
	if (g_codepage[0] == 0)
	{
		if (setlocale(LC_CTYPE, ""))
		{
			STRNCPY(g_codepage, nl_langinfo(CODESET), sizeof(g_codepage));
		}
		else
		{
			STRNCPY(g_codepage, DEFAULT_CODEPAGE, sizeof(g_codepage));
		}
	}
#endif

	if (g_hostname[0] == 0)
	{
		if (gethostname(fullhostname, sizeof(fullhostname)) == -1)
		{
			error("could not determine local hostname, use -n\n");
			return 1;
		}

		p = strchr(fullhostname, '.');
		if (p != NULL)
			*p = 0;

		STRNCPY(g_hostname, fullhostname, sizeof(g_hostname));
	}

	if (g_keymapname[0] == 0)
	{
		if (locale && xkeymap_from_locale(locale))
		{
			fprintf(stdout, "Autoselected keyboard map %s\n", g_keymapname);
		}
		else
		{
			STRNCPY(g_keymapname, "en-us", sizeof(g_keymapname));
		}
	}
	if (locale)
		xfree(locale);


	if (prompt_password && read_password(&password))
		flags |= RDP_LOGON_AUTO;

	if (g_title[0] == 0)
	{
		strcpy(g_title, "rdesktop - ");
		strncat(g_title, server, sizeof(g_title) - sizeof("rdesktop - "));
	}

#ifdef RDP2VNC
	rdp2vnc_connect(server, flags, domain, &password, shell, directory);
	return 0;
#else

	if (!ui_init())
		return 1;

#ifdef WITH_RDPSND
	if (g_rdpsnd)
	{
		if (!rdpsnd_init(rdpsnd_optarg))
		{
			warning("Initializing sound-support failed!\n");
		}
	}
#endif

/* Load external plug-ins */
#ifdef WITH_PLUGIN
	if (plugins != NULL)
	{
		c = 0;
		eluc = 0;

		while (plugins[c].type != NO_PLUGIN)
		{
			switch (plugins[c].type)
			{
#ifdef WITH_WYSE_MMR
				case WYSE_MMR:
					ret = load_wyse_plugins(plugins[c].module);
					break;
#endif
				case ELUSIVA:
					args = NULL;
					i = 1;
					while (plugins[c].args[i] != NULL)
					{
						args = realloc(args,
							(args == NULL ? 0 : (strlen(args) + 1)) + 1 +
							strlen(plugins[c].args[i + 1]));
						args[strlen(args) + 1] = '\0';
						args[strlen(args)] = ':';
						strcpy(&(args[strlen(args)]), plugins[c].args[i]);
						i++;
					}
					if (args == NULL)
						args = strdup("");
					init_external_addin(plugins[c].args[0], plugins[c].module,
						args, &addin_data[eluc]);
					if (addin_data[eluc].pid != 0)
					{
						eluc++;
						ret = True;
					}
					else
					{
						error("Failed to initialise add-in [%s]\n",
							plugins[c].module);
						return 1;
					}
					break;
				default:
					error("Module type is not supported!\n");
					ret = False;
					break;
			}
			if (ret == False)
				return 1;
			c++;
		}
	}
#endif
#ifdef WITH_FABULATECH
	if (g_ftusbrdp)
	{
		if (!ftusbrdp_init())
		{
			warning("Initializing usb-over-rdp failed!\n");
		}
	}
#endif /* WITH_FABULATECH */

	if (g_lspci_enabled)
		lspci_init();

	rdpdr_init();

	while (run_count < 2 && continue_connect)	/* add support for Session Directory; only reconnect once */
	{
		g_licence_issued = False;
		window= show_wait_window(server);
		if (run_count == 0)
		{
			if (!rdp_connect(server, flags, domain, &password, shell, directory, True))
			{
				hide_wait_window(window);
				return 1;
			}
		}
		else if (!rdp_connect(server, flags, domain, &password, shell,directory, False))
		{
			hide_wait_window(window);
			return 1;
		}
		hide_wait_window(window);

		/* By setting encryption to False here, we have an encrypted login 
		   packet but unencrypted transfer of other packets */
		if (!g_packet_encryption)
			g_encryption = False;


		DEBUG(("Connection successful.\n"));

		if (run_count == 0)
			if (!ui_create_window())
				continue_connect = False;

		if (continue_connect)
			rdp_main_loop(&deactivated, &ext_disc_reason);

		if ((g_shutdown == False) && (g_is_connected == True))
		{
			DEBUG(("Disconnecting...\n"));
			rdp_disconnect(False);
		}

		if ((g_redirect == True) && (run_count == 0))	/* Support for Session Directory */
		{
			/* reset state of major globals */
			rdesktop_reset_state();

			STRNCPY(domain, g_redirect_domain, sizeof(domain));
			STRNCPY(g_username, g_redirect_username, sizeof(g_username));
			if(g_redirect_password.len > 0) {
				password.bytes = g_redirect_password.bytes;
				password.len = g_redirect_password.len;
				password.toUTF16 = g_redirect_password.toUTF16;
			}
			STRNCPY(server, g_redirect_server, sizeof(server));
			if ((strlen(domain) > 0) && (strlen(g_username) > 0) &&
				(password.len > 0))
				flags |= RDP_LOGON_AUTO;
#ifdef WITH_SCARD
			/* fix session redirect with smart card. Thanks to "Bernhardt, Stefan" <stefan.bernhardt@siemens.com> */
			/* Check for indicator for Smartcard logon - is this the one need ? */
			if (g_redirect_flags & 0x00000040)
			{
				DEBUG_SCARD(("REDIRECT assigning next session uses smartcard"));
				flags |= 0x00040000; /* tell next server that PIN is in password field */
			};
#endif

			g_redirect = False;
		}
		else
		{
			continue_connect = False;
			ui_destroy_window();
			break;
		}

		run_count++;
	}

	cache_save_state();
	ui_deinit();

#ifdef WITH_PLUGIN
	/* Send a SIGUSR1 to all addins to close and sleep for a couple of secs to
	 * give them a chance to stop
	 */
	for (c = 0; c < addin_count; c++)
		if (addin_data[c].pid != 0)
		{
			kill(addin_data[c].pid, SIGUSR1);
			sleep(2);
		}
#endif

	if (ext_disc_reason >= 2)
		print_disconnect_reason(ext_disc_reason);

	if (deactivated)
	{
		/* clean disconnect */
		return 0;
	}
	else
	{
		if (ext_disc_reason == exDiscReasonAPIInitiatedDisconnect
		    || ext_disc_reason == exDiscReasonAPIInitiatedLogoff)
		{
			/* not so clean disconnect, but nothing to worry about */
			return 0;
		}
		else
		{
			/* return error */
			return 2;
		}
	}

#endif

}

#ifdef EGD_SOCKET
/* Read 32 random bytes from PRNGD or EGD socket (based on OpenSSL RAND_egd) */
static RD_BOOL
generate_random_egd(uint8 * buf)
{
	struct sockaddr_un addr;
	RD_BOOL ret = False;
	int fd;

	fd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (fd == -1)
		return False;

	addr.sun_family = AF_UNIX;
	memcpy(addr.sun_path, EGD_SOCKET, sizeof(EGD_SOCKET));
	if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) == -1)
		goto err;

	/* PRNGD and EGD use a simple communications protocol */
	buf[0] = 1;		/* Non-blocking (similar to /dev/urandom) */
	buf[1] = 32;		/* Number of requested random bytes */
	if (write(fd, buf, 2) != 2)
		goto err;

	if ((read(fd, buf, 1) != 1) || (buf[0] == 0))	/* Available? */
		goto err;

	if (read(fd, buf, 32) != 32)
		goto err;

	ret = True;

      err:
	close(fd);
	return ret;
}
#endif

/* Generate a 32-byte random for the secure transport code. */
void
generate_random(uint8 * random)
{
	struct stat st;
	struct tms tmsbuf;
	SSL_MD5 md5;
	uint32 *r;
	int fd, n;

	/* If we have a kernel random device, try that first */
	if (((fd = open("/dev/urandom", O_RDONLY)) != -1)
	    || ((fd = open("/dev/random", O_RDONLY)) != -1))
	{
		n = read(fd, random, 32);
		close(fd);
		if (n == 32)
			return;
	}

#ifdef EGD_SOCKET
	/* As a second preference use an EGD */
	if (generate_random_egd(random))
		return;
#endif

	/* Otherwise use whatever entropy we can gather - ideas welcome. */
	r = (uint32 *) random;
	r[0] = (getpid()) | (getppid() << 16);
	r[1] = (getuid()) | (getgid() << 16);
	r[2] = times(&tmsbuf);	/* system uptime (clocks) */
	gettimeofday((struct timeval *) &r[3], NULL);	/* sec and usec */
	stat("/tmp", &st);
	r[5] = st.st_atime;
	r[6] = st.st_mtime;
	r[7] = st.st_ctime;

	/* Hash both halves with MD5 to obscure possible patterns */
	ssl_md5_init(&md5);
	ssl_md5_update(&md5, random, 16);
	ssl_md5_final(&md5, random);
	ssl_md5_update(&md5, random + 16, 16);
	ssl_md5_final(&md5, random + 16);
}

/* malloc; exit if out of memory */
void *
xmalloc(int size)
{
	void *mem = malloc(size);
	if (mem == NULL)
	{
		error("xmalloc %d\n", size);
		exit(1);
	}
	return mem;
}

/* Exit on NULL pointer. Use to verify result from XGetImage etc */
void
exit_if_null(void *ptr)
{
	if (ptr == NULL)
	{
		error("unexpected null pointer. Out of memory?\n");
		exit(1);
	}
}

/* strdup */
char *
xstrdup(const char *s)
{
	char *mem = strdup(s);
	if (mem == NULL)
	{
		perror("strdup");
		exit(1);
	}
	return mem;
}

/* realloc; exit if out of memory */
void *
xrealloc(void *oldmem, size_t size)
{
	void *mem;

	if (size == 0)
		size = 1;
	mem = realloc(oldmem, size);
	if (mem == NULL)
	{
		error("xrealloc %ld\n", size);
		exit(1);
	}
	return mem;
}

/* free */
void
xfree(void *mem)
{
	free(mem);
}

/* report an error */
void
error(char *format, ...)
{
	va_list ap;

	fprintf(stderr, "ERROR: ");

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
}

/* report a warning */
void
warning(char *format, ...)
{
	/* workaround: if exit code is not 0, this messages will be shown! */

	va_list ap;

	/* fprintf(stderr, "WARNING: "); */
	fprintf(stdout, "WARNING: ");

	va_start(ap, format);
	/* vfprintf(stderr, format, ap); */
	vfprintf(stdout, format, ap);
	va_end(ap);
}

/* report an unimplemented protocol feature */
void
unimpl(char *format, ...)
{
	va_list ap;

	fprintf(stderr, "NOT IMPLEMENTED: ");

	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
}

/* produce a hex dump */
void
hexdump(unsigned char *p, unsigned int len)
{
	unsigned char *line = p;
	int i, thisline, offset = 0;

	while (offset < len)
	{
		printf("%04x ", offset);
		thisline = len - offset;
		if (thisline > 16)
			thisline = 16;

		for (i = 0; i < thisline; i++)
			printf("%02x ", line[i]);

		for (; i < 16; i++)
			printf("   ");

		for (i = 0; i < thisline; i++)
			printf("%c", (line[i] >= 0x20 && line[i] < 0x7f) ? line[i] : '.');

		printf("\n");
		offset += thisline;
		line += thisline;
	}
}

/*
  input: src is the string we look in for needle.
  	 Needle may be escaped by a backslash, in
	 that case we ignore that particular needle.
  return value: returns next src pointer, for
  	succesive executions, like in a while loop
	if retval is 0, then there are no more args.
  pitfalls:
  	src is modified. 0x00 chars are inserted to
	terminate strings.
	return val, points on the next val chr after ins
	0x00

	example usage:
	while( (pos = next_arg( optarg, ',')) ){
		printf("%s\n",optarg);
		optarg=pos;
	}

*/
char *
next_arg(char *src, char needle)
{
	char *nextval;
	char *p;
	char *mvp = 0;

	/* EOS */
	if (*src == (char) 0x00)
		return 0;

	p = src;
	/*  skip escaped needles */
	while ((nextval = strchr(p, needle)))
	{
		mvp = nextval - 1;
		/* found backslashed needle */
		if (*mvp == '\\' && (mvp > src))
		{
			/* move string one to the left */
			while (*(mvp + 1) != (char) 0x00)
			{
				*mvp = *(mvp + 1);
				mvp++;
			}
			*mvp = (char) 0x00;
			p = nextval;
		}
		else
		{
			p = nextval + 1;
			break;
		}

	}

	/* more args available */
	if (nextval)
	{
		*nextval = (char) 0x00;
		return ++nextval;
	}

	/* no more args after this, jump to EOS */
	nextval = src + strlen(src);
	return nextval;
}


void
toupper_str(char *p)
{
	while (*p)
	{
		if ((*p >= 'a') && (*p <= 'z'))
			*p = toupper((int) *p);
		p++;
	}
}


RD_BOOL
str_startswith(const char *s, const char *prefix)
{
	return (strncmp(s, prefix, strlen(prefix)) == 0);
}


/* Split input into lines, and call linehandler for each
   line. Incomplete lines are saved in the rest variable, which should
   initially point to NULL. When linehandler returns False, stop and
   return False. Otherwise, return True.  */
RD_BOOL
str_handle_lines(const char *input, char **rest, str_handle_lines_t linehandler, void *data)
{
	char *buf, *p;
	char *oldrest;
	size_t inputlen;
	size_t buflen;
	size_t restlen = 0;
	RD_BOOL ret = True;

	/* Copy data to buffer */
	inputlen = strlen(input);
	if (*rest)
		restlen = strlen(*rest);
	buflen = restlen + inputlen + 1;
	buf = (char *) xmalloc(buflen);
	buf[0] = '\0';
	if (*rest)
		STRNCPY(buf, *rest, buflen);
	strncat(buf, input, inputlen);
	p = buf;

	while (1)
	{
		char *newline = strchr(p, '\n');
		if (newline)
		{
			*newline = '\0';
			if (!linehandler(p, data))
			{
				p = newline + 1;
				ret = False;
				break;
			}
			p = newline + 1;
		}
		else
		{
			break;

		}
	}

	/* Save in rest */
	oldrest = *rest;
	restlen = buf + buflen - p;
	*rest = (char *) xmalloc(restlen);
	STRNCPY((*rest), p, restlen);
	xfree(oldrest);

	xfree(buf);
	return ret;
}

/* Execute the program specified by argv. For each line in
   stdout/stderr output, call linehandler. Returns false on failure. */
RD_BOOL
subprocess(char *const argv[], str_handle_lines_t linehandler, void *data)
{
	pid_t child;
	int fd[2];
	int n = 1;
	char output[256];
	char *rest = NULL;

	if (pipe(fd) < 0)
	{
		perror("pipe");
		return False;
	}

	if ((child = fork()) < 0)
	{
		perror("fork");
		return False;
	}

	/* Child */
	if (child == 0)
	{
		/* Close read end */
		close(fd[0]);

		/* Redirect stdout and stderr to pipe */
		dup2(fd[1], 1);
		dup2(fd[1], 2);

		/* Execute */
		execvp(argv[0], argv);
		perror("Error executing child");
		_exit(128);
	}

	/* Parent. Close write end. */
	close(fd[1]);
	while (n > 0)
	{
		n = read(fd[0], output, 255);
		output[n] = '\0';
		str_handle_lines(output, &rest, linehandler, data);
	}
	xfree(rest);

	return True;
}


/* not all clibs got ltoa */
#define LTOA_BUFSIZE (sizeof(long) * 8 + 1)

char *
l_to_a(long N, int base)
{
	static char ret[LTOA_BUFSIZE];

	char *head = ret, buf[LTOA_BUFSIZE], *tail = buf + sizeof(buf);

	register int divrem;

	if (base < 36 || 2 > base)
		base = 10;

	if (N < 0)
	{
		*head++ = '-';
		N = -N;
	}

	tail = buf + sizeof(buf);
	*--tail = 0;

	do
	{
		divrem = N % base;
		*--tail = (divrem <= 9) ? divrem + '0' : divrem + 'a' - 10;
		N /= base;
	}
	while (N);

	strcpy(head, tail);
	return ret;
}


/* This code only works on x86 architecture, because there is no differentiation
 * between big and little endian!
 */
int
load_licence(unsigned char **data)
{
	char *path;
	int fd;
	unsigned char buffer[64];
	uint32 i, len, length, tmp;

	path = (char *) xmalloc(strlen(IGEL_LICENSE_PATH) + strlen(g_hostname) + sizeof("/licence."));
	sprintf(path, "%s/licence.%s", IGEL_LICENSE_PATH, g_hostname);

	fd = open(path, O_RDONLY);
	if (fd == -1)
		return -1;

	if (read(fd, buffer, sizeof( uint32)) != sizeof( uint32))
		return -1;
	
	/* read Scope, Company Name and Product ID */
	for (i= 0; i < 3; i++) {
		if (read(fd, (unsigned char*)&length, sizeof( uint32)) != sizeof( uint32))
			return -1;
		while (length > 0) {
			len= ( length > 64 ? 64 : length);
			if ((tmp= read(fd, buffer, len)) != len)
				return -1;
			length-= len;
		}
	}

	/* read license */
	if (read(fd, (unsigned char*)&length, sizeof( uint32)) != sizeof( uint32))
		return -1;
	*data= (uint8*)xmalloc( length);
	length= read(fd, *data, length);

	close(fd);
	xfree(path);
	return length;
}

void
save_licence(unsigned char *data, int length)
{
	char *path /*, *tmppath */;
	int fd;

	path = (char *) xmalloc(strlen(IGEL_LICENSE_PATH) + strlen(g_hostname) + sizeof("/licence."));

	/* sprintf(path, "%s/.rdesktop", home); */
	if ((mkdir(IGEL_LICENSE_PATH, 0700) == -1) && errno != EEXIST)
	{
		perror(IGEL_LICENSE_PATH);
		return;
	}

	/* write licence to licence.hostname.new, then atomically rename to licence.hostname */

	/* sprintf(path, "%s/.rdesktop/licence.%s", home, g_hostname); */
	sprintf(path, "%s/licence.%s", IGEL_LICENSE_PATH, g_hostname);
	/*
	tmppath = (char *) xmalloc(strlen(path) + sizeof(".new"));
	strcpy(tmppath, path);
	strcat(tmppath, ".new");

	fd = open(tmppath, O_WRONLY | O_CREAT | O_TRUNC, 0600);
	*/
	fd = open(path, O_WRONLY | O_CREAT | O_TRUNC, 0600);
	if (fd == -1)
	{
		/* perror(tmppath); */
		perror(path);
		return;
	}

	if (write(fd, data, length) != length)
	{
		/*
		perror(tmppath);
		unlink(tmppath);
		*/
		perror(path);
		unlink(path);
	}
	/*
	else if (rename(tmppath, path) == -1)
	{
		perror(path);
		unlink(tmppath);
	}
	*/

	close(fd);
	/* xfree(tmppath); */
	xfree(path);
}

/* Create the bitmap cache directory */
RD_BOOL
rd_pstcache_mkdir(void)
{
	char *home;
	char bmpcache_dir[256];

	home = getenv("HOME");

	if (home == NULL)
		return False;

	sprintf(bmpcache_dir, "%s/%s", home, ".rdesktop");

	if ((mkdir(bmpcache_dir, S_IRWXU) == -1) && errno != EEXIST)
	{
		perror(bmpcache_dir);
		return False;
	}

	sprintf(bmpcache_dir, "%s/%s", home, ".rdesktop/cache");

	if ((mkdir(bmpcache_dir, S_IRWXU) == -1) && errno != EEXIST)
	{
		perror(bmpcache_dir);
		return False;
	}

	return True;
}

/* open a file in the .rdesktop directory */
int
rd_open_file(char *filename)
{
	char *home;
	char fn[256];
	int fd;

	home = getenv("HOME");
	if (home == NULL)
		return -1;
	sprintf(fn, "%s/.rdesktop/%s", home, filename);
	fd = open(fn, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
	if (fd == -1)
		perror(fn);
	return fd;
}

/* close file */
void
rd_close_file(int fd)
{
	close(fd);
}

/* read from file*/
int
rd_read_file(int fd, void *ptr, int len)
{
	return read(fd, ptr, len);
}

/* write to file */
int
rd_write_file(int fd, void *ptr, int len)
{
	return write(fd, ptr, len);
}

/* move file pointer */
int
rd_lseek_file(int fd, int offset)
{
	return lseek(fd, offset, SEEK_SET);
}

/* do a write lock on a file */
RD_BOOL
rd_lock_file(int fd, int start, int len)
{
	struct flock lock;

	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start = start;
	lock.l_len = len;
	if (fcntl(fd, F_SETLK, &lock) == -1)
		return False;
	return True;
}

#ifdef WITH_PLUGIN
/* Initialise external addin */
void
init_external_addin(char *addin_name, char *addin_path, char *args, ADDIN_DATA *addin_data)
{
	char *p;
	char *current_arg;
	char *argv[256];
	char argv_buffer[256][256];
	int i;
	int readpipe[2], writepipe[2];
	pid_t child = 0;

	/* Initialise addin structure */
	memset(addin_data, 0, sizeof(ADDIN_DATA));

	/* To through the list of args, addin each to argv */
	argv[0] = addin_path;
	i = 1;
	p = current_arg = args;
	while (current_arg != NULL && current_arg[0] != '\0')
	{
		p = next_arg(p, ':');
		if (p != NULL && *p != '\0')
			*(p - 1) = '\0';
		strcpy(argv_buffer[i], current_arg);
		argv[i] = argv_buffer[i];
		i++;
		current_arg = p;
	}
	argv[i] = NULL;

	/* Create pipes */
	if (pipe(readpipe) < 0 || pipe(writepipe) < 0)
	{
		perror("pipes for addin");
		return;
	}

	/* Fork process */
	if ((child = fork()) < 0)
	{
		perror("fork for addin");
		return;
	}

	/* Child */
	if (child == 0)
	{
		/* Set stdin and stdout of child to relevant pipe ends */
		dup2(writepipe[0], 0);
		dup2(readpipe[1], 1);

		/* Close all fds as they are not needed now */
		close(readpipe[0]);
		close(readpipe[1]);
		close(writepipe[0]);
		close(writepipe[1]);
		execvp((char*)argv[0], (char**)argv);
		perror("Error executing child");
		_exit(128);
	}
	else
	{
		strcpy(addin_data->name, addin_name);
		/* Close child end fd's */
		close(readpipe[1]);
		close(writepipe[0]);
		addin_data->pipe_read = readpipe[0];
		addin_data->pipe_write = writepipe[1];
		addin_data->vchannel = channel_register( addin_name,
			CHANNEL_OPTION_INITIALIZED |
			CHANNEL_OPTION_SHOW_PROTOCOL |
			CHANNEL_OPTION_ENCRYPT_RDP |
			CHANNEL_OPTION_COMPRESS_RDP,
			addin_callback);
		if (!addin_data->vchannel)
		{
			perror("Channel register failed");
			return;
		}
		else
			addin_data->pid = child;
	}
}

/* Find an external add-in registration by virtual channel name */
void
lookup_addin( char *name, pid_t *pid, int *pipe_read, int *pipe_write)
{
	int i;

	*pid = 0;

	for (i = 0; i < addin_count; i++)
		if (!strcmp(name, addin_data[i].name))
		{
			*pid = addin_data[i].pid;
			*pipe_read = addin_data[i].pipe_read;
			*pipe_write = addin_data[i].pipe_write;
			break;
		}
}
#endif

#if defined(WITH_PLUGIN) && defined(WITH_WYSE_MMR)
Window
get_rdp_window()
{
	return g_wnd;
}

int
load_wyse_plugins(char *module)
{
	int ret = False;
	void *handle = NULL;
	int (*rdp_wyse_init)(RDP_SESSION *, char *);

	if ((handle = dlopen(module, RTLD_LAZY)))
	{
		if ((rdp_wyse_init = dlsym(handle, "rdp_wyse_init")))
			ret = rdp_wyse_init(&rdp_session, g_title);
		else
		{
			error("load_wyse_plugins() failed: %s\n", dlerror());
			dlclose(handle);
		}
	}
	else
		error("load_wyse_plugins() failed: %s\n", dlerror());
	
	return ret;
}
#endif

