/*
   rdesktop: A Remote Desktop Protocol client.
   Master include file
   Copyright (C) Matthew Chapman 1999-2007

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef _WIN32
#define WINVER 0x0400
#include <windows.h>
#include <winsock.h>
#include <time.h>
#define DIR int
#else
#include <dirent.h>
#include <sys/time.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif
#endif
#include <limits.h>		/* PATH_MAX */
#if defined(WITH_PLUGIN) && defined(WITH_WYSE_MMR)
#include <X11/Xlib.h>
#endif

#define VERSION "1.6.0.22"
/*
1.6.0.22: If user credentials are given, a session redirection does not longer
          overwrite them if the single-use password value is empty.
1.6.0.21: Session redirection bug has been solved.
1.6.0.20: The "Connecting to" window has been centered to the middle of the
          screen.
1.6.0.19: X errors with cursors could be solved by checking the hotspot x and y
          values. If they are outside of the cursor, they will be set to 0.
1.6.0.18: Session redirection support for Windows Server 2008 and Windows Server
          2008 R2.
1.6.0.17: Correct conversion of UTF-8 strings to UTF-16 strings.
1.6.0.16: If sound mapping version 6 is used, all frames will be written to
          ALSA to avoid a buffer underrrun.
1.6.0.15: Number of channels has been increased to support all extension.
          A bug has been solved which always places the window at -1, -1 if the
          WM class has been used to place the window.
1.6.0.14: Workaround for temporary files which are renamed later. E.g. Excel
          2003 uses this. The file will be generated with lower cases, but
          renamed with upper cases. So we check this and try to rename the
          the file by using the modified name.
1.6.0.13: Support for Multi-Monitor extension (requires Windows Server 2008R2
          or Windows 7). Window class is not longer required, because -M does
          the same. Requires Xinerama.
*/

#ifdef WITH_PLUGIN
#define MAX_ADDINS 20
#endif

enum {
	RDP4= 0,
	RDP5,
	RDP6,
	NUM_RDP
};

#define WAIT_WND_MSG "Connecting to "
#define WAIT_WND_WIDTH 300
#define WAIT_WND_HEIGHT 70

#ifdef WITH_DEBUG
#define DEBUG(args)	printf args;
#else
#define DEBUG(args)
#endif

#ifdef WITH_DEBUG_KBD
#define DEBUG_KBD(args) printf args;
#else
#define DEBUG_KBD(args)
#endif

#ifdef WITH_DEBUG_RDP5
#define DEBUG_RDP5(args) printf args;
#else
#define DEBUG_RDP5(args)
#endif

#ifdef WITH_DEBUG_CLIPBOARD
#define DEBUG_CLIPBOARD(args) printf args;
#else
#define DEBUG_CLIPBOARD(args)
#endif

#ifdef WITH_DEBUG_SOUND
#define DEBUG_SOUND(args) printf args;
#else
#define DEBUG_SOUND(args)
#endif

#ifdef WITH_DEBUG_CHANNEL
#define DEBUG_CHANNEL(args) printf args;
#else
#define DEBUG_CHANNEL(args)
#endif

#ifdef WITH_DEBUG_SCARD
#define DEBUG_SCARD(args) printf args;
#else
#define DEBUG_SCARD(args)
#endif

#define STRNCPY(dst,src,n)	{ strncpy(dst,src,n-1); dst[n-1] = 0; }

#ifndef MIN
#define MIN(x,y)		(((x) < (y)) ? (x) : (y))
#endif

#ifndef MAX
#define MAX(x,y)		(((x) > (y)) ? (x) : (y))
#endif

/* timeval macros */
#ifndef timerisset
#define timerisset(tvp)\
         ((tvp)->tv_sec || (tvp)->tv_usec)
#endif
#ifndef timercmp
#define timercmp(tvp, uvp, cmp)\
        ((tvp)->tv_sec cmp (uvp)->tv_sec ||\
        (tvp)->tv_sec == (uvp)->tv_sec &&\
        (tvp)->tv_usec cmp (uvp)->tv_usec)
#endif
#ifndef timerclear
#define timerclear(tvp)\
        ((tvp)->tv_sec = (tvp)->tv_usec = 0)
#endif

/* If configure does not define the endianess, try
   to find it out */
#if !defined(L_ENDIAN) && !defined(B_ENDIAN)
#if __BYTE_ORDER == __LITTLE_ENDIAN
#define L_ENDIAN
#elif __BYTE_ORDER == __BIG_ENDIAN
#define B_ENDIAN
#else
#error Unknown endianness. Edit rdesktop.h.
#endif
#endif /* B_ENDIAN, L_ENDIAN from configure */

/* No need for alignment on x86 and amd64 */
#if !defined(NEED_ALIGN)
#if !(defined(__x86__) || defined(__x86_64__) || \
      defined(__AMD64__) || defined(_M_IX86) || \
      defined(__i386__))
#define NEED_ALIGN
#endif
#endif

#include "parse.h"
#include "constants.h"
#include "types.h"

#ifndef MAKE_PROTO
#include "proto.h"
#endif

#ifdef WITH_PLUGIN
enum {
	NO_PLUGIN= 0,
	ELUSIVA,
#ifdef WITH_WYSE_MMR
	WYSE_MMR,
#endif
	NUM_TYPES
};

typedef struct {
	int	type;
	char *module;
	char **args;
} PLUGIN;
#endif

#if defined(WITH_PLUGIN) && defined(WITH_WYSE_MMR)
#ifndef __INTL_H
#define __INTL_H
#include <libintl.h>
#define TEXTDOMAIN "RDESKTOP"
#undef _
#define _(s) dgettext(TEXTDOMAIN,s)
#endif
typedef struct {
	Window (*get_rdp_window)();
	VCHANNEL *(*channel_register)(char *, uint32, void (*) (STREAM));
	STREAM (*channel_init)(VCHANNEL *, uint32);
	void (*channel_send)(STREAM, VCHANNEL *);
	void (*channel_process)(STREAM, uint16);
	void (*wave_out_close)(void);
} RDP_SESSION;
#endif
